"""
.. module:: PL
   :synopsis: Represents the propositional logics.

.. moduleauthor:: Alberto Casagrande <acasagrande@units.it>
"""

from .language import *
from ..language import LNot

from .parser import Parser
